# -*- coding: UTF-8 -*-

import os, re, traceback
from common import logger
from upload.converter.shtrihPlugins.shtrihformat import ShtrihFormatter
import xml.etree.ElementTree as ET

errorObjectsCount = 0

class ShtrihParser:
    def __init__(self, node):
        self.log = logger.getUploadLogger("Node", node.id)
        self.delimiter = node.currentConverter.delimiter
        self.cashierName = node.currentConverter.cashierName
        self.hotKeyByBarcode = node.currentConverter.hotKeyByBarcode
        self.requireSaleRestrict = node.currentConverter.requireSaleRestrict
        self.inserter = getattr(__import__('upload.inserter.%s' % node.inserter.name, fromlist='Inserter'), 'Inserter')(node)
        self.packSize = node.pack_size
        self.sprMode = ''
        self.currentFile = None
        self.objects = []
        self.commandRe = re.compile('\A([%s]+|\<[a-zA-Z0-9_]+)' % re.escape(u'#?!@%^()~&+-*/{}[]’«{\$–=№|'))
        self.formatter = ShtrihFormatter(self.log, self.delimiter, self.cashierName, self.hotKeyByBarcode, self.requireSaleRestrict)
        self.goodsAttrCommands = {}
        self.errorList = []
        self.filePosition = 0
        self.currentFileSize = 0

    def parse(self, data):
        """
        Метод парсинга всех файлов
        data - список файлов
        Возвращает результат вставки (aif-объектов, список ошибок)
        """
        if not self.inserter.start():
            self.log.error(u'Не удалось запустить Inserter')
            return False, self.errorList
        try:
            for filePath in data:
                self.parseSingleFile(filePath)
        except Exception as e:
            self.log.error(e + (u'(позиция в файле %s)' % self.filePosition))
            self.log.error(traceback.format_exc())
            self.errorList.append(e)
            return False, self.errorList
        result = self.inserter.stop()
        global errorObjectsCount
        errorObjectsCount = self.inserter.invalidObjectsCount()
        return result, self.errorList

    def parseSingleFile(self, filePath):
        '''
        Парсинг данных в файле
        В предыдущей версии конвертера мы считали, что в одном файле может быть несколько секций с данными
        filePath - путь до файла
        '''
        self.currentFile = open(filePath, 'rb+') # 'rb' для корректной работы функции file.tell() в Windows
        self.filePosition = 0
        self.currentFileSize = os.fstat(self.currentFile.fileno()).st_size
        while self.readSection(): # читать секции пока они есть
            pass

    def readSection(self):
        '''
        Чтение секции. Секция: данные между '##@@&&' и следующим '##@@&&' или концом файла
        False, если конец файла
        Возвращает True, если секция прочитана успешно,
        Исключения: Exception - если в начале секции нет '##@@&&'
        '''
        self.sprMode = 'RPL' # По умолчанию для каждой секции включен режим "Замены"
        line = self.nextLine()
        if line.startswith('##@@&&'):
            self.loadObjects(self.getKeyFromLine(line[6:]), self.packSize) # чтение флагов загрузки в заголовке секции после '##@@&&'
            sectionFlagPosition = self.currentFile.tell()
            if self.readSectionFlag():
                self.readSectionData()
                self.loadObjects(self.getDeferredObjects(), 0) # загрузка aif-объектов отложенных команд
                self.updateSectionFlag(sectionFlagPosition)
            else:
                self.skipSection()
        elif not line:
            return False # пустая строка - считаем что конец файла
        else:
            raise Exception(u'Неверный формат файла загрузки %s' % self.currentFile.name)
        return True

    def readSectionFlag(self):
        '''
        Чтение флага секции
        Возвращает True, если есть флаг загрузки,
        False, если данные загружать не нужно
        Исключения: Exception - если флаг загрузки не найден
        '''
        line = self.nextLine()
        if line == '@': # если вторая строка секции '@', то секцию загружать не нужно
            return False
        elif line == '#':
            return True
        else:
            raise Exception(u'Флаг загрузки не найден %s' % self.currentFile.name)

    def readSectionData(self):
        '''
        Чтение данных секции. Получение и загрузка aif-объектов
        '''
        self.loadObjects(self.findSectionKeys(), self.packSize)
        while True:
            position = self.currentFile.tell()
            line = self.nextLine()
            if not line:
                return # конец файла
            elif line.startswith('##@@&&'): # текущая секция закончилась и началась новая; возвращаем каретку на начало секции
                self.currentFile.seek(position)
                return
            elif line.startswith('$$$'): # все флаги($$$CLR, $$$ADD) уже былы загружены методом findSectionKeys()
                continue
            elif line == '@@@EOF' or line == '@@@ENABLECHECK': # игнорируем непонятные флаги
                continue
            else:
                self.loadObjects(self.getData(line), self.packSize)

    def getData(self, line):
        '''
        Функция получения aif-объектов для каждой пары (команда, данные)
        line - текущая строка
        Возвращает список aif-объектов
        '''
        aifObjects = []
        for command, line in self.processCommands(line):
            aifObjects.extend(self.getAifObjects(command, line))
        return aifObjects

    def processCommands(self, line):
        '''
        Получение команды и её анализ на отложенность
        Отложенные команды обрабатываются после загрузки всех данных секции
        line - текущая строка
        Метод возвращает список пар (команда, данные)
        '''
        regularCommands = []
        command, line = self.getFullCommand(line)
        if command.startswith('<'):
            try:
                xmlRoot = ET.fromstring(line.encode('utf-8'))
                for xmlItem in xmlRoot:
                    if xmlItem.tag == 'goods_attr':
                        tmcCode = xmlItem.attrib['id'].strip()
                        tmcAttr = xmlItem.attrib['attr_id'].strip()
                        tmcAttrValue = xmlItem.text.strip()
                        if tmcCode in self.goodsAttrCommands.keys():
                            self.goodsAttrCommands[tmcCode].update({tmcAttr: tmcAttrValue})
                        else:
                            self.goodsAttrCommands[tmcCode] = {tmcAttr: tmcAttrValue}
                    else:
                        regularCommands.append((command, ET.tostring(xmlItem, encoding='utf-8')))
            except Exception:
                self.log.error(traceback.format_exc())
                self.errorList.append(u'Ошибка парсинга xml-команды(%s) в %s(позиция в файле %s)' % (line, self.currentFile.name, self.filePosition))
                self.log.error(self.errorList[-1])
        else:
            regularCommands.append((command, line))
        return regularCommands

    def getFullCommand(self, line):
        '''
        Получение команды и строки с данными
        Команда может быть:
            однострочной, не xml - обычная Штрих-команда
            многострочной - необходимо последовательно считать несколько строк с данными, чтобы обеспечить создание aif-объекта
            xml - в общем случае многострочная команда
        Поскольку одна команда формата Штрих может занимать несколько строк,
        то функция читает строки до конца многострочной команды
        line - текущая строка
        Метод возвращает:
            для однострочной не xml-команды: (команда, строка без команды)
            для многострочной команды: (команда, строка без команды состоящая из объединения всех строк многострочной команды с '\n' в качестве разделителя)
            для xml-команды: <rows><возможно, несколько xml-команд></rows>
        Исключения: Exception - при ошибке структуры многострочной команды
        '''
        command = self.getCommand(line)
        if command.startswith('<'):
            return command, self.getXmlCommands(command, line)
        elif command.startswith('{@'):
            # считаем что разрезы представляют из себя блок данных вида:
            # '{@...dataA...\n'
            # '{#...dataB1...\n'
            # '...'
            # '{#...dataBN...\n'
            # '{№...dataC...\n'
            # парсим таким образом, чтобы функция вернула команду '{@' и ОДНУ строку, данные которой разделены '\n': '{@...dataA...\n...dataB1...\n...dataBN...\n...dataC...'
            resultCommand = '{@'
            resultLine = line[len(command):]
            while True:
                line = self.nextLine()
                command = self.getCommand(line)
                if command == '{#':
                    resultLine += '\n' + line[len(command):]
                elif command == u'{№':
                    resultLine += '\n' + line[len(command):]
                    break
                else:
                    raise Exception(u'Ошибка структуры справочника загрузки. Ожидалась команда "{№" или "{#"')
            return resultCommand, resultLine
        elif command.startswith('&'): 
            # в команде добавления кассира после & могут присутствовать спец. символы *, # и т.д., 
            # тогда команда определяется как &*, &#, а таких команд нет в списке
            resultCommand = '&'
            resultLine = line[len(resultCommand):]
            return resultCommand, resultLine
        elif len(command): # обычная однострочная команда Штрих
            return command, line[len(command):]

        # поиск кастомных команд, добавленных Artix
        if line[0].isdigit() or (line[0].isalpha() and len(line.split(self.delimiter))) == 17:
            line = self.addInventSprMode(line, self.sprMode)
        return line[0], line[1:]

    def getCommand(self, line):
        '''
        Метод получения документированной команды из текущей строки
        line - текущая строка
        Возвращает символьную команду или пустую строку, если команда кастомная
        '''
        commandMatch = self.commandRe.findall(line)
        return commandMatch[0] if len(commandMatch) else ''

    def getXmlCommands(self, command, line):
        '''
        Получение xml-команд из файла загрузки
        Чтение происходит пока не встретим закрывающийся тэг в конце полученной строки
        Может так получиться, что в одной строке будут 2 или более xml-команды,
        поэтому чтобы xml.etree.ElementTree распарсил нужно обернуть данные в корневой элемент
        command - xml-команда вида '<команда'
        line - текущая строка
        Возвращает строку вида <rows><возможно, несколько xml-команд></rows>
        '''
        closingTag = '</' + command[1:] + '>' # закрывающий тэг xml-команды
        if line.endswith('/>') or line.endswith(closingTag):
            return '<rows>' + line + '</rows>'
        multilineCommand = line
        while True:
            line = self.nextXmlLine()
            if not line:
                raise Exception(u'Ошибка чтения данных в %s, команда %s' % (self.currentFile.name, command))
            multilineCommand += line
            if line.strip().endswith(closingTag):
                break
        return '<rows>' + multilineCommand + '</rows>'

    def getAifObjects(self, command, line):
        '''
        Получение aif-объектов из команды и данных в формате Штрих
        command - команда
        line - строка с данными
        Возвращает список aif-объектов
        '''
        aifObjects = []
        if command in self.formatter.formatDescription:
            try:
                result = self.formatter.formatDescription[command](line)
                aifObjects.extend(result) if isinstance(result, list) else aifObjects.append(result)
            except Exception as e:
                self.log.error(traceback.format_exc())
                self.errorList.append(u'Ошибка в команде %s, поцизия в файле %s. %s' % (command, self.filePosition, e))
                self.log.error(self.errorList[-1])
        else:
            self.log.warn(u'Команда "%s" не поддержана' % command)
        return filter(None, aifObjects)

    def updateSectionFlag(self, sectionFlagPosition):
        '''
        Изменение флага загрузки в секции после получения данных
        sectionFlagPosition - позиция в файле с флагом загрузки(указывает на начало строки,
        поэтому если в файле символ '#' находится не в начале строки, то будет некорректное поведение)
        '''
        tmpPos = self.currentFile.tell()
        self.currentFile.seek(sectionFlagPosition)
        self.currentFile.write('@'.encode('utf8'))
        self.currentFile.seek(tmpPos)

    def addInventSprMode(self, line, mode):
        '''
        Получение строки с комадной добавления товарной позиции
        line - строка добавления товарной позиции в формате Штрих
        mode - режим загрузки товарной позиции
        Возвращает модифицированную строку
        '''
        return '0' + line + self.delimiter + mode

    def getKeyFromLine(self, line):
        '''
        Получение различных ключей($$$...) из строки
        line - текущая строка
        Возвращает список aif-объектов удаления данных из таблиц
        '''
        result = []
        flagsRegularExpression = re.compile('(\{[^\}]*\})')
        if line.startswith('$$$'):
            if line.startswith('$$$CLR'):
                for flag in flagsRegularExpression.findall(line):
                    if flag in self.formatter.clearCommands:
                        clrCmd = self.formatter.clearCommands[flag]
                        result.extend(clrCmd) if isinstance(clrCmd, list) else result.append(clrCmd)
                if not self.formatter.clearCommands['{NO_TOV}'] in result:
                    result.append(self.formatter.clearCommands['{NO_TOV}'])
                else:
                    result.remove(self.formatter.clearCommands['{NO_TOV}'])
            elif line.startswith('$$$ADD'):
                self.sprMode = 'ADD'
                self.formatter.setSprMode(self.sprMode)
        return result

    def findSectionKeys(self):
        '''
        Поиск всех ключей, начинающихся с $$$ в секции на которой установлена каретка
        Возвращает список aif-объектов удаления данных из таблиц
        '''
        result = []
        position = self.currentFile.tell()
        line = self.nextLine()
        while not line.startswith('##@@&&') and line:
            if line.startswith('$$$'):
                result += self.getKeyFromLine(line)
            line = self.nextLine()
        self.currentFile.seek(position)
        return result

    def skipSection(self):
        '''
        Пропуск секции путём считывания строк до '##@@&&' или конца файла
        При этом каретка устанавливается на конец строки перед строкой с '##@@&&'
        '''
        position = self.currentFile.tell()
        line = self.nextLine()
        while not line.startswith('##@@&&') and line:
            position = self.currentFile.tell()
            line = self.nextLine()
        self.currentFile.seek(position)

    def loadObjects(self, objects, size):
        '''
        Добавление объектов в inserter
        objects - список объектов, которые нужно загрузить,
        size - порог срабатывания загрузки
        '''
        self.objects.extend(objects)
        if len(self.objects) >= size:
            self.inserter.add_package(self.objects)
            self.objects = [] # объекты загружены - очищаем список

    def getDeferredObjects(self):
        '''
        Получение aif-объектов отложенных команд и очистка списка отложенных команд
        Возвращает список aif-объектов
        '''
        result = self.formatter.formatDescription['<goods_attr'](self.goodsAttrCommands)
        self.goodsAttrCommands.clear()
        return filter(None, result)

    def nextLine(self):
        '''
        Получение следующей не пустой строки текущего файла загрузки
        Также перед считыванием строки получает текущую позицию файла
        Возвращает декодированную строку без пробельных символов
        '''
        self.filePosition = self.currentFile.tell()
        while self.currentFile.tell() < self.currentFileSize:
            line = self.currentFile.readline().strip().decode('cp1251')
            if len(line):
                return line
        return ''

    def nextXmlLine(self):
        '''
        Получение следующей не пустой строки текущего файла загрузки
        Также перед считыванием строки получает текущую позицию файла
        Возвращает декодированную оригинальную строку
        '''
        self.filePosition = self.currentFile.tell()
        while self.currentFile.tell() < self.currentFileSize:
            line = self.currentFile.readline().decode('cp1251')
            if len(line):
                return line
        return ''

def general_method(workDirectory, data, node):
    '''
    Создание парсера и начало парсинга файлов
    workDirectory - рабочая директория, куда копируются файлы со справочниками
    data - массив путей до файлов со справочниками
    node - информация об узле для загрузки справочников
    Возвращает результат вставки (aif-объектов, список ошибок)
    '''
    return ShtrihParser(node).parse(data)

def invalidObjectsCount():
    '''
    Возвращает количество объектов не прошедших загрузку
    '''
    return errorObjectsCount
